"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const url_1 = require("url");
const apollo_server_core_1 = require("apollo-server-core");
const apollo_server_env_1 = require("apollo-server-env");
const getRequestBody = (req) => new Promise(resolve => {
    let body;
    req.on('data', (chunk) => {
        if (!body)
            body = chunk;
        else
            body = Buffer.concat([body, chunk]);
    });
    req.on('end', () => {
        resolve(body.toString());
    });
});
const createQueryRequest = (req) => {
    const method = req.method;
    if (!method)
        throw new Error('Missing method');
    const url = req.url;
    if (!url)
        throw new Error('Missing url');
    const rHeaders = new apollo_server_env_1.Headers();
    for (const [name, value] of Object.entries(req.headers)) {
        if (!value)
            continue;
        if (Array.isArray(value)) {
            value.forEach(v => rHeaders.append(name, v));
        }
        else {
            rHeaders.append(name, value);
        }
    }
    return new apollo_server_env_1.Request(url, {
        headers: rHeaders,
        method
    });
};
function graphqlHttp(options) {
    if (!options) {
        throw new Error('Apollo Server requires options.');
    }
    if (arguments.length > 1) {
        throw new Error(`Apollo Server expects exactly one argument, got ${arguments.length}`);
    }
    return async (req, res) => {
        const method = req.method;
        if (!method)
            throw new Error('Missing method');
        const url = req.url;
        if (!url)
            throw new Error('Missing url');
        apollo_server_core_1.runHttpQuery([req, res], {
            method,
            options: options,
            query: req.method === 'POST'
                ? JSON.parse(await getRequestBody(req))
                : new url_1.URL(url).searchParams,
            request: createQueryRequest(req)
        }).then(({ graphqlResponse, responseInit }) => {
            if (responseInit.headers) {
                for (const [name, value] of Object.entries(responseInit.headers)) {
                    res.setHeader(name, value);
                }
            }
            res.write(graphqlResponse);
            res.end();
        }, (error) => {
            if ('HttpQueryError' !== error.name) {
                throw error;
            }
            if (error.headers) {
                for (const [name, value] of Object.entries(error.headers)) {
                    res.setHeader(name, value);
                }
            }
            res.statusCode = error.statusCode;
            res.write(error.message);
            res.end();
        });
    };
}
exports.graphqlHttp = graphqlHttp;
//# sourceMappingURL=httpApollo.js.map